#******************************************************************************************************
#  binaryStream.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  01/31/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .streamEncoder import streamEncoder
from .encoding7Bit import encoding7Bit
from . import ByteSize, Validate
from typing import Optional
from uuid import UUID
import sys
import numpy as np

class binaryStream:
    """
    Establishes buffered I/O around a base stream, e.g., a socket.
    """
    
    # Source C# references:
    #     RemoteBinaryStream
    #     BinaryStreamBase

    IOBufferSize = 1420
    ValueBufferSize = 16

    def __init__(self, stream: streamEncoder):
        self.stream = stream
        self.defaultByteOrder = stream.DefaultByteOrder
        self.defaultIsNative = self.defaultByteOrder == sys.byteorder
        self.buffer = bytearray(binaryStream.ValueBufferSize)
        self.receiveBuffer = bytearray(binaryStream.IOBufferSize)
        self.sendBuffer = bytearray(binaryStream.IOBufferSize)
        self.sendLength = 0
        self.receiveLength = 0
        self.receivePosition = 0

    @property
    def SendBufferFreeSpace(self) -> int:
        return binaryStream.IOBufferSize - self.sendLength

    @property
    def ReceiveBufferAvailable(self) -> int:
        return self.receiveLength - self.receivePosition

    def Flush(self):
        if self.sendLength <= 0:
            return

        self.stream.Write(bytes(self.sendBuffer), 0, self.sendLength)
        self.sendLength = 0

    def Read(self, buffer: bytearray, offset: int, count: int) -> int:
        if count <= 0:
            return 0;

        receiveBufferLength = self.ReceiveBufferAvailable

        # Check if there is enough in the receive buffer to handle request
        if count <= receiveBufferLength:
            for i in range(count):
                buffer[offset + i] = self.receiveBuffer[self.receivePosition + i]

            self.receivePosition += count
            return count

        originalCount = count

        # Empty existing receive buffer
        if receiveBufferLength > 0:
            for i in range(receiveBufferLength):
                buffer[offset + i] = self.receiveBuffer[self.receivePosition + i]

            self.receivePosition = 0
            self.receiveLength = 0
            offset += receiveBufferLength
            count -= receiveBufferLength
        
        # If more than 100 bytes remain, skip receive buffer 
        # and copy directly to the destination
        if count > 100:
            # Loop since socket reads can return partial results
            while count > 0:
                receiveBufferLength = self.stream.Read(buffer, offset, count)

                if receiveBufferLength == 0:
                    raise RuntimeError("End of stream")

                offset += receiveBufferLength
                count -= receiveBufferLength            
        else:
            # With fewer than 100 bytes requested, fill receive buffer 
            # then copy to destination
            prebufferLength = binaryStream.IOBufferSize
            self.receiveLength = 0

            while self.receiveLength < count:
                receiveBufferLength = self.stream.Read(self.receiveBuffer, self.receiveLength, prebufferLength)
                
                if receiveBufferLength == 0:
                    raise RuntimeError("End of stream")

                self.receiveLength += receiveBufferLength
                prebufferLength -= receiveBufferLength

            for i in range(count):
                buffer[offset + i] = self.receiveBuffer[i]

            self.receivePosition = count

        return originalCount

    def Write(self, buffer: bytes, offset: int, count: int) -> int:
        if self.SendBufferFreeSpace < count:
            self.Flush()

        if count > 100:
            self.Flush()
            self.stream.Write(buffer, offset, count)
        else:
            for i in range(count):
                self.sendBuffer[self.sendLength + i] = buffer[offset + i]
            
            self.sendLength += count

        return count

    def ReadAll(self, buffer: bytearray, position: int, length: int):
        """
        Reads all of the provided bytes. Will not return prematurely, continues
        to execute `Read` operation until the entire `length` has been read.
        """
        Validate.Parameters(buffer, position, length)

        while length > 0:
            bytesRead = self.Read(buffer, position, length)
            
            if bytesRead == 0:
                raise RuntimeError("End of stream")

            length -= bytesRead
            position += bytesRead

    def ReadBytes(self, count: int) -> bytes:
        buffer = self.buffer if count <= binaryStream.ValueBufferSize else bytearray(count)
        self.ReadAll(buffer, 0, count)
        return bytes(buffer[0:count])

    def ReadBuffer(self) -> bytes:
        return self.ReadBytes(self.Read7BitUInt32())

    def ReadString(self) -> str:
        return self.ReadBuffer().decode("utf-8")
    
    def ReadGuid(self) -> UUID:
        return UUID(bytes_le=self.ReadBytes(16))

    def Read7BitInt32(self) -> np.int32:
        return encoding7Bit.ReadInt32(self.ReadByte)
        
    def Read7BitUInt32(self) -> np.uint32:
        return encoding7Bit.ReadUInt32(self.ReadByte)

    def Read7BitInt64(self) -> np.int64:
        return encoding7Bit.ReadInt64(self.ReadByte)

    def Read7BitUInt64(self) -> np.uint64:
        if self.receivePosition <= self.receiveLength - 9:
            stream = streamEncoder(self.__sendBufferRead, self.__sendBufferWrite)
            return stream.Read7BitUInt64()

        return encoding7Bit.ReadUInt64(self.ReadByte)

    def ReadByte(self) -> np.uint8:
        size = ByteSize.UINT8

        if self.receivePosition < self.receiveLength:
            value = self.receiveBuffer[self.receivePosition]
            self.receivePosition += size
            return np.uint8(value)

        self.ReadAll(self.buffer, 0, size)
        return self.buffer[0]

    def WriteBuffer(self, value: bytes) -> int:
        count = len(value)
        return self.Write7BitUInt32(count) + self.Write(value, 0, count)

    def WriteString(self, value: str) -> int:
        return self.WriteBuffer(value.encode("utf-8"))
    
    def WriteGuid(self, value: UUID) -> int:        
        return self.Write(value.bytes_le, 0, 16)

    def Write7BitInt32(self, value: np.int32) -> int:
        return encoding7Bit.WriteInt32(self.WriteByte, value)

    def Write7BitUInt32(self, value: np.uint32) -> int:
        return encoding7Bit.WriteUInt32(self.WriteByte, value)

    def Write7BitInt64(self, value: np.int64) -> int:
        return encoding7Bit.WriteInt64(self.WriteByte, value)
    
    def Write7BitUInt64(self, value: np.uint64) -> int:
        if self.sendLength <= binaryStream.IOBufferSize - 9:
            stream = streamEncoder(self.__sendBufferRead, self.__sendBufferWrite)
            return stream.Write7BitUInt64(value)

        return encoding7Bit.WriteUInt64(self.WriteByte, value)

    def WriteByte(self, value: np.uint8) -> int:
        size = ByteSize.UINT8

        if self.sendLength < binaryStream.IOBufferSize:
            self.sendBuffer[self.sendLength] = value
            self.sendLength += size
            return size
        
        self.buffer[0] = value
        return self.Write(self.buffer, 0, size)

    def ReadBoolean(self) -> bool:
        return self.ReadByte() != 0

    def WriteBoolean(self, value: bool) -> int:
        if value:
            self.WriteByte(1)
        else:
            self.WriteByte(0)

        return 1

    def __readInt(self, size: int, dtype: np.dtype, byteorder: Optional[str]) -> int:
        if not (byteorder is None and self.defaultIsNative) and byteorder != sys.byteorder:
            dtype = dtype.newbyteorder()

        if self.receivePosition <= self.receiveLength - size:
            value = np.frombuffer(self.receiveBuffer[self.receivePosition:self.receivePosition + size], dtype)[0]
            self.receivePosition += size
            return value

        self.ReadAll(self.buffer, 0, size)
        return np.frombuffer(self.buffer[0:size], dtype)[0]

    def __writeInt(self, size: int, value: int, signed: bool, byteorder: Optional[str]) -> int:
        buffer = int(value).to_bytes(size, self.defaultByteOrder if byteorder is None else byteorder, signed=signed)

        if self.sendLength <= binaryStream.IOBufferSize - size:
            for i in range(size):
                self.sendBuffer[self.sendLength + i] = buffer[i]

            self.sendLength += size
            return size

        for i in range(size):
            self.buffer[i] = buffer[i]

        return self.Write(self.buffer, 0, size)

    def ReadInt16(self, byteorder: Optional[str] = None) -> np.int16:
        return self.__readInt(ByteSize.INT16, np.dtype(np.int16), byteorder)

    def WriteInt16(self, value: np.int16, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.INT16, value, True, byteorder)

    def ReadUInt16(self, byteorder: Optional[str] = None) -> np.uint16:
        return self.__readInt(ByteSize.UINT16, np.dtype(np.uint16), byteorder)

    def WriteUInt16(self, value: np.uint16, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.UINT16, value, False, byteorder)

    def ReadInt32(self, byteorder: Optional[str] = None) -> np.int32:
        return self.__readInt(ByteSize.INT32, np.dtype(np.int32), byteorder)

    def WriteInt32(self, value: np.int32, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.INT32, value, True, byteorder)

    def ReadUInt32(self, byteorder: Optional[str] = None) -> np.uint32:
        return self.__readInt(ByteSize.UINT32, np.dtype(np.uint32), byteorder)

    def WriteUInt32(self, value: np.uint32, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.UINT32, value, False, byteorder)

    def ReadInt64(self, byteorder: Optional[str] = None) -> np.int64:
        return self.__readInt(ByteSize.INT64, np.dtype(np.int64), byteorder)

    def WriteInt64(self, value: np.int64, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.INT64, value, True, byteorder)

    def ReadUInt64(self, byteorder: Optional[str] = None) -> np.uint64:
        return self.__readInt(ByteSize.UINT64, np.dtype(np.uint64), byteorder)

    def WriteUInt64(self, value: np.uint64, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.UINT64, value, False, byteorder)

    def __sendBufferRead(self, length: int) -> bytes:
        buffer = self.buffer if length <= binaryStream.ValueBufferSize else bytearray(length)

        for i in range(length):
            buffer[i] = self.receiveBuffer[self.receivePosition + i]

        self.receivePosition += length
        return bytes(buffer[0:length])

    def __sendBufferWrite(self, buffer: bytes) -> int:
        length = len(buffer)

        for i in range(length):
            self.sendBuffer[self.sendLength + i] = buffer[i]

        self.sendLength += length
        return length



#******************************************************************************************************
#  treeStream.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/05/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from abc import ABC, abstractmethod
from typing import TypeVar, Generic

TKey = TypeVar('TKey')
TValue = TypeVar('TValue')

class treeStream(ABC, Generic[TKey, TValue]):
    """
    Represents a stream of SNAPdb key/value pairs.
    """

    def __init__(self):
        self.eos = False
        self.disposed = False

    @property
    def EOS(self) -> bool:
        """
        Gets flag indicating if the end of the stream has been read or class has been disposed.
        """
        return self.eos

    @property
    def IsDisposed(self) -> bool:
        """
        Gets flag indicating if class has been disposed.
        """
        return self.disposed

    @property
    def IsAlwaysSequential(self) -> bool:
        """
        Gets flag indicating if the stream is always in sequential order. Do not return true unless
        it is guaranteed that the data read from this stream is sequential.
        """
        return False

    @property        
    def NeverContainsDuplicates(self) -> bool:
        """
        Gets flag indicating if the stream will never return duplicate keys. Do not return true unless
        it is guaranteed that the data read from this stream will never contain duplicates.
        """
        return False

    def Read(self, key: TKey, value: TValue) -> bool:
        """
        Advances the stream to the next value.
        """
        if self.eos or not self.ReadNext(key, value):
            self.EndOfStreamReached()
            return False

        return True

    def EndOfStreamReached(self):
        """
        Occurs when the end of the stream has been reached. The default behavior is to call `Dispose`.
        """
        self.Dispose()

    def SetEOS(self, value: bool):
        """
        Resets the EOS flag.
        """
        if self.disposed and not value:
            raise ValueError(f"Stream {type(self).__name__} has already been disposed")

        self.eos = value

    @abstractmethod
    def ReadNext(self, key: TKey, value: TValue) -> bool:
        """
        Advances the stream to the next value.  If before the beginning of the stream, advances to
        the first value. Returns `True` if the advance was successful; otherwise, `False` if the
        end of the stream was reached.
        """

    def Dispose(self):
        """
        Derived classes should not override `Dispose` method. Instead, derived classes
        should override `Disposing` method for custom class shutdown operations.
        """
        if self.disposed:
            return

        try:
            self.Disposing()
        finally:
            self.eos = True
            self.disposed = True
  
    def Disposing(self):
        """
        Derived classes should override this method for custom class shutdown operations.
        """
        self.eos = True



#******************************************************************************************************
#  openHistorian.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  01/31/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .historianInstance import historianInstance
from .historianKey import historianKey
from .historianValue import historianValue
from .metadataCache import metadataCache
from snapDB.snapConnection import snapConnection
from snapDB.encodingDefinition import encodingDefinition
from gsf.streamEncoder import streamEncoder
from gsf.binaryStream import binaryStream
from gsf import override
from typing import Optional, Callable
from enum import IntEnum
from time import time, sleep
import errno
import socket
import gzip
import numpy as np

class ServerCommand(IntEnum):
    # Meta data refresh command.
    METADATAREFRESH = 0x01
    # Define operational modes for subscriber connection.
    DEFINEOPERATIONALMODES = 0x06

class ServerResponse(IntEnum):
    # Command succeeded response.
    SUCCEEDED = 0x80
    # Command failed response.
    FAILED = 0x81

class historianConnection(snapConnection[historianKey, historianValue]):
    """
    Defines API functionality for connecting to an openHistorian instance then
    reading and writing measurement data from the instance.

    This class is an instance of the `snapConnection` implemented for the
    openHistorian `historianKey` and `historianValue` SNAPdb types.
    """

    def __init__(self, hostAddress: str):
        super().__init__(hostAddress, historianKey(), historianValue())
        self.metadata: Optional[metadata] = None

    @override
    def OpenInstance(self, instanceName: str, definition: Optional[encodingDefinition] = None) -> historianInstance:
        return super().OpenInstance(instanceName, definition)

    @property
    def Metadata(self) -> Optional[metadataCache]:
        return self.metadata

    def RefreshMetadata(self, sttpPort: int = 7175, logOutput: Optional[Callable[[str], None]] = None) -> int:
        """
        Requests updated metadata from openHistorian connection.
        """
        sttpSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)        
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        stream = binaryStream(streamEncoder((lambda length: sttpSocket.recv(length)), (lambda buffer: sttpSocket.send(buffer)), "big"))
        
        logOutput = (lambda value: print(value)) if logOutput is None else logOutput
        logOutput("Requesting metadata from openHistorian...")
        opStart = time()

        # Using STTP connection to get metadata only (no subscription), hence the following operational modes:
        # OperationalModes.CompressMetadata | CompressionModes.GZip | OperationalEncoding.UTF8 | (OperationalModes.VersionMask & 1U) 
        operationalModes = 0x80000221

        try:
            sttpSocket.connect((self.HostIPAddress, sttpPort))

            # Establish operational modes for STTP connection
            stream.WriteInt32(5) # Payload aware buffer length
            stream.WriteByte(ServerCommand.DEFINEOPERATIONALMODES)
            stream.WriteUInt32(operationalModes)

            # Request metadata refresh
            stream.WriteInt32(1) # Payload aware buffer length
            stream.WriteByte(ServerCommand.METADATAREFRESH)
            stream.Flush()

            while True:
                # Get payload aware buffer length
                bufferLength = stream.ReadInt32()
                
                # Get server response
                responseCode = ServerResponse(stream.ReadByte())
                commandCode = ServerCommand(stream.ReadByte())
                length = 0 if bufferLength < 3 else stream.ReadInt32()
                
                # Read response payload
                buffer = historianConnection.ReadBytes(stream, length)

                # Other commands can come spontaneously, like NoOp,
                # only interested in MetadataRefresh response
                if commandCode != ServerCommand.METADATAREFRESH:
                    logOutput(f"Ignoring response 0x{responseCode.value:x} received for command 0x{commandCode.value:x}...")
                    continue
            
                if responseCode == ServerResponse.SUCCEEDED:                
                    logOutput(f"Received {length:,} bytes of metadata in {(time() - opStart):.2f} seconds. Decompressing...")
                    opStart = time()

                    try:
                        # Decompress full metadata response XML
                        buffer = gzip.decompress(buffer)
                    except Exception as ex:
                        raise RuntimeError(f"Failed to decompress metadata: {ex}")

                    logOutput(f"Decompressed {len(buffer):,} bytes of metadata in {(time() - opStart):.2f} seconds. Parsing...")
                    opStart = time()

                    try:
                        # Parse and cache received metadata XML
                        self.metadata = metadataCache(buffer.decode("utf-8"))
                    except Exception as ex:
                        raise RuntimeError(f"Failed to parse metadata: {ex}")

                    measurementRecordCount = len(self.metadata.MeasurementRecords)
                    deviceRecordCount = len(self.metadata.DeviceRecords)
                    phasorRecordCount = len(self.metadata.PhasorRecords)
                    recordCount = measurementRecordCount + deviceRecordCount + phasorRecordCount
                    
                    logOutput(f"Parsed {recordCount:,} metadata records in {(time() - opStart):.2f} seconds.")
                    logOutput(f"    Discovered:")
                    logOutput(f"        {measurementRecordCount:,} measurement records")
                    logOutput(f"        {deviceRecordCount:,} device records, and")
                    logOutput(f"        {phasorRecordCount:,} phasor records")

                    # Return total number of metadata records
                    return recordCount
                else:
                    raise RuntimeError(f"Failure code received in response to STTP metadata refresh request: {buffer.decode('utf-8')}")
        finally:
            sttpSocket.close()

    @staticmethod
    def ReadBytes(stream: binaryStream, length: int) -> bytes:
        buffer = bytearray(length)
        position = 0

        while length > 0:
            bytesRead = stream.Read(buffer, position, length)
            
            if bytesRead == 0:
                raise RuntimeError("End of stream")

            length -= bytesRead
            position += bytesRead

        return bytes(buffer)

